'''pytest configuration'''
from ctypes.util import find_library


def pytest_generate_tests(metafunc):
    '''Run tests with the `pcre` fixture twice; Once with `pcre=False`, another
    time with `pcre=True` if the library is present.'''
    if find_library('pcre2-8') and 'pcre' in metafunc.fixturenames:
        metafunc.parametrize('pcre', [False, True])



'''The command line utility for ChromaTerm'''
# pylint: disable=import-outside-toplevel
import argparse
import atexit
import io
import os
import re
import select
import signal
import sys
from ctypes.util import find_library

import yaml

from .core import Color, Config, Palette, Rule
from .__version__ import __version__

# Possible locations of the config file, without the extension. Most-specific
# locations lead in the list.
CONFIG_LOCATIONS = [
    '~/.chromaterm',
    os.getenv('XDG_CONFIG_HOME', '~/.config') + '/chromaterm/chromaterm',
    '/etc/chromaterm/chromaterm',
]

# Maximum chuck size per read
READ_SIZE = 8192

# Sequences upon which ct will split during processing (ECMA 035 and 048):
# * new lines, vertical spaces, form feeds;
# * private control functions, C1 set (excluding control strings);
# * independent control functions (\e#), SCS (G0 through G3 sets);
# * CSI (excluding SGR); and
# * control strings (DSC, SOS, OSC, PM, APC).
SPLIT_RE = re.compile(
    br'(\r\n?|[\n\v\f]|'
    br'\x1b[\x30-\x4f\x51-\x57\x59-\x5a\x5c\x60-\x7e]|'
    br'\x1b[\x23\x28-\x2b\x2d-\x2f][\x20-\x7e]|'
    br'\x1b\x5b[\x30-\x3f]*[\x20-\x2f]*[\x40-\x6c\x6e-\x7e]|'
    br'\x1b[\x50\x58\x5d\x5e\x5f][^\x07\x1b]*(?:\x07|\x1b\x5c)?)')

# Control strings that have arbitrary lengths
ANSI_CONTROL_STRINGS_START = (b'\x1b\x50', b'\x1b\x58', b'\x1b\x5d',
                              b'\x1b\x5e', b'\x1b\x5f')
ANSI_CONTROL_STRINGS_END = (b'\x07', b'\x1b\x5c')


def args_init(args=None):
    '''Returns the parsed arguments (an instance of argparse.Namespace).

    Args:
        args (list): A list of program arguments, Defaults to sys.argv.
    '''
    formatter = lambda prog: argparse.HelpFormatter(prog, max_help_position=30)
    parser = argparse.ArgumentParser(formatter_class=formatter)
    parser.epilog = 'For more info, go to https://github.com/hSaria/ChromaTerm.'

    parser.add_argument('program',
                        metavar='program ...',
                        nargs='?',
                        help='run a program with anything after it used as '
                        'arguments')

    parser.add_argument('arguments',
                        nargs=argparse.REMAINDER,
                        help=argparse.SUPPRESS)

    parser.add_argument('-b',
                        '--benchmark',
                        action='store_true',
                        help='at exit, print rule usage statistics')

    parser.add_argument('-c',
                        '--config',
                        metavar='file',
                        help='override config file location (default: ~/'
                        '.chromaterm.yml)')

    parser.add_argument('-r',
                        '--reload',
                        action='store_true',
                        help='reload the config of all CT instances')

    parser.add_argument('-R',
                        '--rgb',
                        action='store_true',
                        help='use RGB colors (default: detect support, '
                        'fallback to xterm-256)')

    parser.add_argument('-v',
                        '--version',
                        action='version',
                        version=f'%(prog)s {__version__}')

    parser.add_argument('--pcre',
                        action='store_true',
                        help="(advanced) use PCRE2 instead of Python's RE"
                        if find_library('pcre2-8') else argparse.SUPPRESS)

    args = parser.parse_args(args=args)

    # Detect rgb support if not specified
    args.rgb = args.rgb or os.getenv('COLORTERM') in ('truecolor', '24bit')

    return args


def eprint(*args, **kwargs):
    '''Prints a message to stderr.'''
    # Use \r\n to move to beginning of the new line in raw mode
    print('ct:', *args, end='\r\n', file=sys.stderr, **kwargs)


def get_default_config_location():
    '''Returns the first location in `CONFIG_LOCATIONS` that points to a file,
    defaulting to the first item in the list.'''
    resolve = lambda x: os.path.expanduser(os.path.expandvars(x))

    for location in CONFIG_LOCATIONS:
        for extension in ['.yml', '.yaml']:
            path = resolve(location + extension)

            if os.path.isfile(path):
                return path

    # No file found; default to the most-specific location
    return resolve(CONFIG_LOCATIONS[0] + '.yml')


def get_wait_duration(buffer, min_wait=1 / 256, max_wait=1 / 8):
    '''Returns the duration (float) to wait for more data before the last chunk
    of the received data can be processed independently.

    ChromaTerm cannot determine if it's processing data faster than input rate or
    if the input has finished. Therefore, ChromaTerm waits before processing the
    last chunk in the buffer. The waiting is cancelled if data becomes available.
    1/256 is smaller (shorter) than the fastest key repeat (1/255 second). 1/8th
    of a second is generally enough to account for the latency of command
    processing or a remote connection.

    Args:
        buffer (bytes): The incoming data that was processed.
        min_wait (float): Lower-bound on the returned duration.
        max_wait (float): Upper-bound on the returned duration.
    '''
    # New lines indicate long output; relax the wait duration
    if b'\n' in buffer:
        return max_wait

    return min_wait


def load_config(config, data, rgb=False, pcre=False):
    '''Reads configuration from a YAML-based string, formatted like so:
    palette:
      red: "#ff0000"
      blue: "#0000ff"

    rules:
    - description: My first rule
        regex: Hello
        color: b.red
    - description: My second rule is specfic to groups
        regex: W(or)ld
        color:
        0: f#321321
        1: bold

    Any errors are printed to stderr.

    Args:
        config (chromaterm.Config): The instance to which data is loaded.
        data (str): A string containg YAML data.
        rgb (bool): Whether the terminal is RGB-capable or not.
    '''
    try:
        data = yaml.safe_load(data) or {}
    except yaml.YAMLError as exception:
        eprint('Parse error:', exception)
        return

    palette = data.get('palette') if isinstance(data, dict) else None
    rules = data.get('rules') if isinstance(data, dict) else None

    if palette is not None:
        if not isinstance(palette, dict):
            eprint('Parse error: "palette" is not a dictionary')
            return

        palette = parse_palette(palette)

        if not isinstance(palette, Palette):
            eprint(palette)
            return

    if rules is None:
        eprint('Parse error: "rules" list not found in configuration')
        return

    if not isinstance(rules, list):
        eprint('Parse error: "rules" is not a list')
        return

    config.rules.clear()

    for rule in rules:
        rule = parse_rule(rule, palette=palette, rgb=rgb, pcre=pcre)

        if isinstance(rule, Rule):
            config.rules.append(rule)
        else:
            eprint(rule)

    # Put the non-overlapping (exclusive=True) rules at the top
    config.rules = sorted(config.rules, key=lambda x: not x.exclusive)


def parse_palette(data):
    '''Returns an instance of chromaterm.Palette if parsed correctly. Otherwise,
    a string with the error message is returned.

    Args:
        data (dict): A dictionary representing the color palette, formatted
            according to `chromaterm.__main__.load_config`.
    '''
    palette = Palette()

    try:
        for name, value in data.items():
            palette.add_color(name, value)
    except (TypeError, ValueError) as exception:
        return f'Error on palette color {repr(name)}: {repr(value)}: {exception}'

    return palette


def parse_rule(data, palette=None, rgb=False, pcre=False):
    '''Returns an instance of chromaterm.Rule if parsed correctly. Otherwise, a
    string with the error message is returned.

    Args:
        data (dict): A dictionary representing the rule, formatted according to
            `chromaterm.__main__.load_config`.
        palette (chromaterm.Palette): A palette to resolve colors.
        rgb (bool): Whether the terminal is RGB-capable or not.
    '''
    if not isinstance(data, dict):
        return f'Rule {repr(data)} not a dictionary'

    description = data.get('description')
    exclusive = data.get('exclusive', False)
    regex = data.get('regex')
    color = data.get('color')
    rule_repr = f'Rule(regex={repr(str(regex)[:30])})'

    if description:
        rule_repr = rule_repr[:-1] + f', description={repr(description)})'

    if not isinstance(color, dict):
        color = {0: color}

    try:
        for group, value in color.items():
            color[group] = Color(value, palette=palette, rgb=rgb)

        return Rule(regex, color, description, exclusive, pcre)
    except (TypeError, ValueError) as exception:
        return f'Error on {rule_repr}: {exception}'
    except re.error as exception:
        return f'Error on {rule_repr}: re.error: {exception}'


def process_input(config, data_fd, forward_fd=None, max_wait=None):
    '''Processes input by reading from data_fd, highlighting it using config,
    then printing it to sys.stdout. If forward_fd is not None, any data it has
    will be written (forwarded) into data_fd.

    Args:
        config (chromaterm.Config): Used for highlighting the data.
        data_fd (int): File descriptor or socket to be read and highlighted.
        forward_fd (int): File descriptor or socket to be forwarded into data_fd.
        max_wait (float): The maximum time to wait with no data on either of the
            file descriptors. None will block until at least one ready to be read.
    '''
    # pylint: disable=too-many-branches
    # #93: Avoid BlockingIOError when output to non-blocking stdout is too fast
    try:
        getattr(os, 'set_blocking', lambda *_: None)(sys.stdout.fileno(), True)
    except io.UnsupportedOperation:
        pass

    if isinstance(data_fd, int):
        forward = lambda: os.write(data_fd, os.read(forward_fd, READ_SIZE))
        read = lambda: os.read(data_fd, READ_SIZE)
    else:
        forward = lambda: data_fd.sendall(forward_fd.recv(READ_SIZE))
        read = lambda: data_fd.recv(READ_SIZE)

    fds = [data_fd] if forward_fd is None else [data_fd, forward_fd]
    buffer = b''

    ready_fds = read_ready(*fds, timeout=max_wait)

    while ready_fds:
        # There's some data to forward to the spawned program
        if forward_fd in ready_fds:
            try:
                forward()
            except OSError:
                # Spawned program or stdin closed; don't forward anymore
                fds.remove(forward_fd)

        # Data to be highlighted was received
        if data_fd in ready_fds:
            try:
                data_read = read()
            except OSError:
                data_read = b''

            buffer += data_read

            # Buffer was processed empty and data fd hit EOF
            if not buffer:
                break

            chunks = split_buffer(buffer)

            # Process chunks except for the last one as it might've been cut off
            for data, separator in chunks[:-1]:
                sys.stdout.buffer.write(config.highlight(data) + separator)

            # Flush as `read_ready` might delay the flush at the bottom
            sys.stdout.flush()

            # Wait is cancelled when the user is typing; relax the wait duration
            if forward_fd is not None:
                wait_duration = get_wait_duration(buffer, min_wait=1 / 64)
            else:
                wait_duration = get_wait_duration(buffer)

            data, separator = chunks[-1]

            # Separator is an incomplete control string; wait for the rest
            if data_read and separator.startswith(ANSI_CONTROL_STRINGS_START) \
                    and not separator.endswith(ANSI_CONTROL_STRINGS_END):
                buffer = data + separator
            # A single character indicates keyboard typing; don't highlight
            # Account for backspaces added by some shells, like zsh
            elif 0 < len(buffer) < 2 + data_read.count(b'\b') * 2:
                sys.stdout.buffer.write(data + separator)
                buffer = b''
            # There's more data; fetch it before highlighting, if buffer isn't full
            elif data_read and len(buffer) <= READ_SIZE \
                    and data_fd in read_ready(*fds, timeout=wait_duration):
                buffer = data + separator
            else:
                sys.stdout.buffer.write(config.highlight(data) + separator)
                buffer = b''

            # Flush as the last chunk might not end with a new line
            sys.stdout.flush()

        ready_fds = read_ready(*fds, timeout=max_wait)


def read_file(location):
    '''Returns the contents of a file or `None` on error. The error is printed
    to stderr.

    Args:
        location (str): The location of the file to be read.
    '''
    if not os.access(location, os.F_OK):
        eprint(f'Configuration file {repr(location)} not found')
        return None

    try:
        with open(location, 'r', encoding='utf-8') as file:
            return file.read()
    except PermissionError:
        eprint(f'Cannot read configuration file {repr(location)} (permission)')
        return None


def read_ready(*fds, timeout=None):
    '''Returns a list of file descriptors that are ready to be read.

    Args:
        *fds (int): Integers that refer to the file descriptors.
        timeout (float): Passed to `select.select`.
    '''
    return [] if not fds else select.select(fds, [], [], timeout)[0]


def signal_chromaterm_instances(sig):
    '''Sends `sig` signal to all other ChromaTerm CLI instances.

    Returns:
        The number of processes reloaded.
    '''
    import psutil

    count = 0
    current_process = psutil.Process()

    for process in psutil.process_iter():
        # Don't reload the current process
        if process.pid == current_process.pid:
            continue

        try:
            # Only compare the first two arguments (Python and script paths)
            if process.cmdline()[:2] == current_process.cmdline()[:2]:
                os.kill(process.pid, sig)
                count += 1
        except (psutil.AccessDenied, psutil.NoSuchProcess):
            pass

    return count


def split_buffer(buffer):
    '''Returns a tuple of tuples in the format of (data, separator). data should
    be highlighted while separator should be printed, unchanged, after data.

    Args:
        buffer (bytes): Data to split using `SPLIT_RE`.
    '''
    chunks = SPLIT_RE.split(buffer)

    # Append an empty separator when chunks end with data
    if chunks[-1]:
        chunks.append(b'')

    # Group all chunks into format of (data, separator)
    return tuple(zip(chunks[0::2], chunks[1::2]))


def main(args=None, max_wait=None, write_default=True):
    '''Command line utility entry point.

    Args:
        args (list): A list of program arguments. Defaults to sys.argv.
        max_wait (float): The maximum time to wait with no data. None will block
            until data is ready to be read.
        write_default (bool): Whether to write the default configuration or not.
            Only written if it doesn't exist already.

    Returns:
        A string indicating status/error. Otherwise, returns None.
    '''
    # pylint: disable=expression-not-assigned
    args = args_init(args)

    if args.reload:
        return f'Processes reloaded: {signal_chromaterm_instances(signal.SIGUSR1)}'

    # Config file wasn't overridden; use default file
    if not args.config:
        args.config = get_default_config_location()

        # Write default config if not there
        if write_default and not os.access(args.config, os.F_OK):
            from . import default_config
            default_config.write_default_config(args.config)

    config = Config(benchmark=args.benchmark)

    # Print benchmark after cleanup in `run_program`
    if args.benchmark:
        atexit.register(config.print_benchmark_results)

    from .platform import unix as platform

    if args.program:
        # ChromaTerm is spawning the program in a pty; stdin is forwarded
        data_fd = platform.run_program([args.program] + args.arguments)
        forward_fd = platform.get_stdin()
    else:
        # Data is being piped into ChromaTerm's stdin; no forwarding needed
        data_fd = platform.get_stdin()
        forward_fd = None

    # Signal handler to trigger reloading the config
    def reload_config_handler(*_):
        config_data = read_file(args.config)

        if config_data:
            load_config(config, config_data, rgb=args.rgb, pcre=args.pcre)

    # Trigger the initial loading
    reload_config_handler()

    # Ignore SIGINT (CTRL+C) and attach reload handler
    signal.signal(signal.SIGINT, signal.SIG_IGN)
    signal.signal(signal.SIGUSR1, reload_config_handler)

    try:
        # Begin processing the data (blocking operation)
        process_input(config, data_fd, forward_fd, max_wait)
    except BrokenPipeError:
        # Surpress the implicit flush that Python runs on exit
        sys.stdout.flush = lambda: None
    finally:
        # Close data_fd to signal to the child process that we're done
        os.close(data_fd) if isinstance(data_fd, int) else data_fd.close()

    return os.wait()[1] >> 8 if args.program else 0



'''chromaterm.platform.unix tests'''
import sys

import chromaterm.platform.unix


def test_get_stdin(monkeypatch):
    '''Ensure stdin is a file descriptor.'''
    monkeypatch.setattr(sys.stdin, 'fileno', lambda: 0)
    assert isinstance(chromaterm.platform.unix.get_stdin(), int)



'''chromaterm.pcre tests'''
import re
from ctypes.util import find_library
from unittest.mock import MagicMock

import pytest

# pylint: disable=protected-access,wrong-import-position

if not find_library('pcre2-8'):  # pragma: no cover
    pytest.skip('skipping PCRE tests; lib not preset', allow_module_level=True)

import chromaterm.pcre


def test_finditer():
    '''Find all matches in data.'''
    regex = chromaterm.pcre.Pattern(b'hello (world)')
    match = list(regex.finditer(b'--- hello world ---'))[0]

    assert match.span(0) == (4, 15)
    assert match.span(1) == (10, 15)


def test_free_memory(monkeypatch):
    '''Free the PCRE `regex` and `match_data` on cleanup.'''
    mock = MagicMock()
    regex = chromaterm.pcre.Pattern(b'hello')
    _match_data = regex._match_data
    _regex = regex._regex

    monkeypatch.setattr(chromaterm.pcre.PCRE2, 'pcre2_match_data_free_8', mock)
    monkeypatch.setattr(chromaterm.pcre.PCRE2, 'pcre2_code_free_8', mock)

    del regex
    assert mock.mock_calls[0][1][0] is _match_data
    assert mock.mock_calls[1][1][0] is _regex


def test_groupindex():
    '''Lookup named groups.'''
    regex = chromaterm.pcre.Pattern(br'(?P<hi>hello)( )(?P<place>world)')

    assert regex.groupindex['hi'] == 1
    assert regex.groupindex.get('place') == 3
    assert regex.groupindex.get('no') is None


def test_re_error():
    '''Invalid regular expression should raise `re.error`.'''
    with pytest.raises(re.error, match=r'position \d+: '):
        chromaterm.pcre.Pattern(b'hello (world')


def test_zero_length_matching():
    '''Ensure that a zero-length matches don't put PCRE into an infinite loop.'''
    regex = chromaterm.pcre.Pattern(br'(?<=hello)\b')
    assert len(list(regex.finditer(b'hello world'))) == 0



'''Unix/macOS functions'''
import atexit
import fcntl
import os
import pty
import signal
import sys
import termios
import threading
import time
import tty

# The frequency to check the child process' `cwd` and update our own
CWD_UPDATE_INTERVAL = 1 / 4


def create_cwd_watcher(pid):
    '''
    Creates a daemon thread that checks the `cwd` of a process, then updates the
    `cwd` of the current process.

    Args:
        pid (int): PID of the process whose `cwd` is checked.
    '''

    def update_cwd():  # pragma: no cover
        # Covered by `test_main_cwd_tracking` but not detected by coverage
        # pylint: disable=import-outside-toplevel
        import psutil

        try:
            child_process = psutil.Process(pid)
        except (psutil.AccessDenied, psutil.NoSuchProcess):
            return

        while True:
            try:
                time.sleep(CWD_UPDATE_INTERVAL)
                os.chdir(child_process.cwd())
            except (OSError, psutil.AccessDenied, psutil.NoSuchProcess):
                pass

    # Daemonized to exit immediately with ChromaTerm
    threading.Thread(target=update_cwd, daemon=True).start()


def get_stdin():
    '''Returns the file descriptor for stdin.'''
    return sys.stdin.fileno()


def run_program(program_args):
    '''Returns a file descriptor of the bidirectional pipe to the spawned program.

    Args:
        program_args (list): A list of program arguments. The first argument is
            the program name/location.
    '''
    try:
        # Set to raw as the pty will be handling any processing; restore at exit
        attributes = termios.tcgetattr(sys.stdin.fileno())
        atexit.register(termios.tcsetattr, sys.stdin.fileno(), termios.TCSANOW,
                        attributes)
        tty.setraw(sys.stdin.fileno(), termios.TCSANOW)
    except termios.error:
        attributes = None

    # openpty, login_tty, then fork
    pid, master_fd = pty.fork()

    # The fork process is replaced with exec or it exits if it hits an exception
    if pid == 0:
        # Update the slave's pty (now on standard fds) attributes
        if attributes:
            termios.tcsetattr(sys.stdin.fileno(), termios.TCSANOW, attributes)

        try:
            os.execvp(program_args[0], program_args)
        except OSError as exception:
            sys.exit(f'{exception.strerror.lower()}: {program_args[0]}')

    # Update the slave's window size as the master is capturing the signals
    def window_resize_handler(*_):
        size = fcntl.ioctl(sys.stdin.fileno(), termios.TIOCGWINSZ, '0000')
        fcntl.ioctl(master_fd, termios.TIOCSWINSZ, size)

    if sys.stdin.isatty():
        signal.signal(signal.SIGWINCH, window_resize_handler)
        window_resize_handler()

    # Some terminals update their titles based on `cwd` (see #94)
    create_cwd_watcher(pid)

    return master_fd



'''chromaterm tests'''
import pytest

import chromaterm

# pylint: disable=too-many-lines

# Aliases to two reset codes used during tests
RESET_BOLD = chromaterm.COLOR_TYPES['bold']['reset']
RESET_ITALIC = chromaterm.COLOR_TYPES['italic']['reset']


def make_sgr(*code_id):
    '''Returns byte-array for an SGR of `code_id` (multiple are concatenated).'''
    return b''.join(b'\x1b[' + x + b'm' for x in code_id)


def test_color_known_colors():
    '''Random known-good color codes.'''
    known_colors = [('644db0', 61), ('18ce9f', 43), ('7318d8', 56),
                    ('5c96f6', 69), ('1ca4c3', 37), ('d9dcc1', 187),
                    ('8bfe8c', 120), ('33f743', 83), ('36d840', 77),
                    ('5baf98', 72), ('e5a667', 179), ('255bfe', 27),
                    ('53e7b4', 79), ('7126d5', 56), ('60c51c', 76),
                    ('bb9465', 137), ('15f502', 46), ('673f0b', 58),
                    ('818fbb', 103), ('38fad2', 86), ('7ba6f1', 111),
                    ('b6bffb', 147), ('cdf107', 190), ('4340ad', 61),
                    ('c9b248', 179), ('903a02', 94), ('084bef', 27),
                    ('a7dca2', 151), ('bf9a28', 136), ('20ca8e', 42),
                    ('670f41', 53), ('94b8e9', 110), ('2c2d27', 235),
                    ('d56561', 167), ('042e6b', 17), ('e7122a', 160),
                    ('3420d6', 56), ('eae8fb', 255), ('04cc5e', 41),
                    ('6a0577', 54), ('6c6e1f', 58), ('aa1ff2', 129),
                    ('8d33db', 98), ('85f7c2', 121), ('b55896', 132),
                    ('254118', 235), ('9c6c95', 132), ('a877ab', 139),
                    ('7e4739', 95), ('4e8867', 65), ('7fbe9a', 108),
                    ('e6da8e', 186), ('c355d7', 134), ('582017', 52),
                    ('93051d', 88), ('3ac6a4', 79), ('fae310', 220),
                    ('08f2d7', 50), ('09fbf1', 51), ('1ab465', 35),
                    ('cf4341', 167), ('042ba4', 19), ('5310e0', 56),
                    ('6d732e', 58), ('6d7cec', 69), ('c428d9', 164),
                    ('4fd7f4', 81), ('09de3e', 41), ('410cfd', 57),
                    ('e41749', 161), ('dc4578', 168), ('ebf705', 190),
                    ('ce22b2', 163), ('ccf0ba', 193), ('a2a315', 142),
                    ('db3a0e', 166), ('8ff3dd', 122), ('866a61', 95),
                    ('25b6b4', 37), ('adbbb1', 249), ('d8272a', 160),
                    ('2c72d8', 26), ('e401df', 164), ('10081b', 233),
                    ('b3bf40', 143), ('0b572e', 22), ('d3a6de', 182),
                    ('71659e', 61), ('260ff7', 21), ('81bb90', 108),
                    ('8d7b5e', 101), ('ebf324', 190), ('1d6925', 22),
                    ('e966c9', 170), ('f0e172', 221), ('7b2f0f', 88),
                    ('fea6d8', 218), ('065694', 24), ('1e2e82', 18),
                    ('ab4e33', 131)]

    for hex_code, xterm256_code in known_colors:
        color = chromaterm.Color(f'f#{hex_code}', rgb=False)
        assert color.color_code == make_sgr(b'38;5;%d' % xterm256_code)


def test_color_decode_sgr_bg():
    '''Background colors and reset are being detected.'''
    for code in [b'48;5;12', b'48;2;1;1;1', b'49', b'101']:
        colors = chromaterm.Color.decode_sgr(make_sgr(code))
        assert len(colors) == 1

        for color_code, _, color_type in colors:
            assert color_code == make_sgr(code)
            assert color_type == 'bg'


def test_color_decode_sgr_fg():
    '''Foreground colors and reset are being detected.'''
    for code in [b'38;5;12', b'38;2;1;1;1', b'39', b'91']:
        colors = chromaterm.Color.decode_sgr(make_sgr(code))
        assert len(colors) == 1

        for color_code, _, color_type in colors:
            assert color_code == make_sgr(code)
            assert color_type == 'fg'


def test_color_decode_sgr_styles_blink():
    '''Blink and its reset are being detected.'''
    for code in [b'5', b'25']:
        colors = chromaterm.Color.decode_sgr(make_sgr(code))
        assert len(colors) == 1

        for color_code, _, color_type in colors:
            assert color_code == make_sgr(code)
            assert color_type == 'blink'


def test_color_decode_sgr_styles_bold():
    '''Bold and its reset are being detected.'''
    for code in [b'1', b'2', b'22']:
        colors = chromaterm.Color.decode_sgr(make_sgr(code))
        assert len(colors) == 1

        for color_code, _, color_type in colors:
            assert color_code == make_sgr(code)
            assert color_type == 'bold'


def test_color_decode_sgr_styles_invert():
    '''Invert and its reset are being detected.'''
    for code in [b'7', b'27']:
        colors = chromaterm.Color.decode_sgr(make_sgr(code))
        assert len(colors) == 1

        for color_code, _, color_type in colors:
            assert color_code == make_sgr(code)
            assert color_type == 'invert'


def test_color_decode_sgr_styles_italic():
    '''Italic and its reset are being detected.'''
    for code in [b'3', b'23']:
        colors = chromaterm.Color.decode_sgr(make_sgr(code))
        assert len(colors) == 1

        for color_code, _, color_type in colors:
            assert color_code == make_sgr(code)
            assert color_type == 'italic'


def test_color_decode_sgr_styles_strike():
    '''Strike and its reset are being detected.'''
    for code in [b'9', b'29']:
        colors = chromaterm.Color.decode_sgr(make_sgr(code))
        assert len(colors) == 1

        for color_code, _, color_type in colors:
            assert color_code == make_sgr(code)
            assert color_type == 'strike'


def test_color_decode_sgr_styles_underline():
    '''Underline and its reset are being detected.'''
    for code in [b'4', b'24']:
        colors = chromaterm.Color.decode_sgr(make_sgr(code))
        assert len(colors) == 1

        for color_code, _, color_type in colors:
            assert color_code == make_sgr(code)
            assert color_type == 'underline'


def test_color_decode_sgr_full_reset():
    '''Full reset detection.'''
    for code in [b'00', b'0', b'']:
        colors = chromaterm.Color.decode_sgr(make_sgr(code))
        assert len(colors) == 1

        for color_code, color_reset, color_type in colors:
            assert color_code == make_sgr(b'0')
            assert color_reset is True
            assert color_type is None


def test_color_decode_sgr_malformed():
    '''Malformed colors.'''
    for code in [b'38;5', b'38;2;1;1', b'38;5;123;38;2;1;1', b'>1;']:
        colors = chromaterm.Color.decode_sgr(make_sgr(code))
        assert len(colors) == 1

        for color_code, color_reset, color_type in colors:
            assert color_code == make_sgr(code)
            assert color_reset is False
            assert color_type is None


def test_color_decode_sgr_split_compound():
    '''Split the a compound SGR into discrete SGR's.'''
    codes = [b'1', b'33', b'40']
    types = ['bold', 'fg', 'bg']

    colors = chromaterm.Color.decode_sgr(make_sgr(b'1;33;40'))
    assert len(colors) == 3

    for index, (color_code, is_reset, color_type) in enumerate(colors):
        assert color_code == make_sgr(codes[index])
        assert is_reset is False
        assert color_type == types[index]


def test_color_decode_sgr_unrecognized():
    '''An SGR that's valid, but the type isn't recognized during decoding.'''
    colors = chromaterm.Color.decode_sgr(make_sgr(b'60'))
    assert len(colors) == 1

    for color_code, color_reset, color_type in colors:
        assert color_code == make_sgr(b'60')
        assert color_reset is False
        assert color_type is None


def test_color_decode_sgr_is_reset():
    '''Colors and styles with `is_reset=True` should always be marked as resets.'''
    for style, code in zip(['fg', 'bg', 'bold'],
                           [b'38;5;12', b'48;5;12', b'1']):
        colors = chromaterm.Color.decode_sgr(make_sgr(code), is_reset=True)
        assert len(colors) == 1

        for color_code, color_reset, color_type in colors:
            assert color_code == make_sgr(code)
            assert color_reset
            assert color_type == style


def test_color_change_color():
    '''Confirm that the color_code is updated when the color or rgb is changed.'''
    color = chromaterm.Color('b#123123', rgb=False)

    # Change from background to foreground
    old_color_code = color.color_code
    color.color = 'f#123123'
    assert color.color_code != old_color_code

    # Toggle rgb
    old_color_code = color.color_code
    color.rgb = not color.rgb
    assert color.color_code != old_color_code


def test_color_duplicate_style():
    '''Duplicate styles are ignored as opposed to added multiple times.'''
    for name, color_type in chromaterm.COLOR_TYPES.items():
        # Only styles have a 'code' key
        if color_type.get('code'):
            color = chromaterm.Color(f'{name} {name}')

            assert color.color == name
            assert color.color_code == color_type['code']
            assert color.color_reset == color_type['reset']


def test_color_palette():
    '''Use a palette color.'''
    palette = chromaterm.Palette()
    palette.add_color('red', '#ff0000')

    color = chromaterm.Color('b.red bold', palette=palette, rgb=False)

    assert color.color == 'b.red bold'
    assert color.color_code == make_sgr(b'48;5;196', b'1')
    assert color.color_reset == make_sgr(b'22', b'49')
    assert color.color_types == [('bg', make_sgr(b'48;5;196')),
                                 ('bold', make_sgr(b'1'))]


def test_color_invalid_value_palette_missing():
    '''Use a palette color, but don't specify a pleatte.'''
    with pytest.raises(ValueError, match='no palette specified'):
        chromaterm.Color('b.red bold')


def test_color_format_color_background():
    '''Background color.'''
    color = chromaterm.Color('b#123123', rgb=False)

    assert color.color == 'b#123123'
    assert color.color_code == make_sgr(b'48;5;235')
    assert color.color_reset == make_sgr(b'49')
    assert color.color_types == [('bg', make_sgr(b'48;5;235'))]
    assert color.rgb is False


def test_color_format_color_foreground():
    '''Foreground color.'''
    color = chromaterm.Color('f#123123', rgb=False)

    assert color.color == 'f#123123'
    assert color.color_code == make_sgr(b'38;5;235')
    assert color.color_reset == make_sgr(b'39')
    assert color.color_types == [('fg', make_sgr(b'38;5;235'))]
    assert color.rgb is False


def test_color_format_color_multiple():
    '''Multiple different color types. Also tests case insensitivity'''
    color = chromaterm.Color('f#123123 b#aBCDef bOLd', rgb=False)

    assert color.color == 'f#123123 b#abcdef bold'
    assert color.color_code == make_sgr(b'38;5;235', b'48;5;153', b'1')
    assert color.color_reset == make_sgr(b'22', b'49', b'39')
    assert color.color_types == [
        ('fg', make_sgr(b'38;5;235')),
        ('bg', make_sgr(b'48;5;153')),
        ('bold', make_sgr(b'1')),
    ]
    assert color.rgb is False


def test_color_format_color_rgb():
    '''RGB color.'''
    color = chromaterm.Color('b#123123', rgb=True)

    assert color.color == 'b#123123'
    assert color.color_code == make_sgr(b'48;2;18;49;35')
    assert color.color_reset == make_sgr(b'49')
    assert color.color_types == [('bg', make_sgr(b'48;2;18;49;35'))]
    assert color.rgb is True


def test_color_format_color_style():
    '''Style color.'''
    styles = {k: v for k, v in chromaterm.COLOR_TYPES.items() if v.get('code')}

    for style in styles:
        color = chromaterm.Color(style, rgb=False)

        assert color.color == style
        assert color.color_code == styles[style]['code']
        assert color.color_reset == styles[style]['reset']
        assert color.color_types == [(style, styles[style]['code'])]
        assert color.rgb is False


def test_color_invalid_type_color():
    '''Color with an int for color.'''
    with pytest.raises(TypeError, match='color must be a string'):
        chromaterm.Color(123)


def test_color_invalid_type_rgb():
    '''Color with an int for rgb.'''
    with pytest.raises(TypeError, match='rgb must be a boolean'):
        chromaterm.Color('bold', rgb=1)


def test_color_invalid_value_background():
    '''Color with an incorrect background color.'''
    with pytest.raises(ValueError, match='invalid color format'):
        chromaterm.Color('b#12312')

    with pytest.raises(ValueError, match='invalid color format'):
        chromaterm.Color('b#12312x')


def test_color_invalid_value_foreground():
    '''Color with an incorrect foreground color.'''
    with pytest.raises(ValueError, match='invalid color format'):
        chromaterm.Color('f#12312')

    with pytest.raises(ValueError, match='invalid color format'):
        chromaterm.Color('f#12312x')


def test_color_invalid_value_style():
    '''Color with an incorrect style.'''
    with pytest.raises(ValueError, match='invalid color format'):
        chromaterm.Color('something')


def test_color_invalid_value_duplicate_background():
    '''Color with multiple background colors.'''
    with pytest.raises(ValueError, match='one foreground and one background'):
        chromaterm.Color('b#123123 b#456456')


def test_color_invalid_value_duplicate_foreground():
    '''Color with multiple background colors.'''
    with pytest.raises(ValueError, match='one foreground and one background'):
        chromaterm.Color('f#123123 f#456456')


def test_palette_add_color():
    '''Confirm a color and its value are added to the palette.'''
    palette = chromaterm.Palette()
    palette.add_color('red', '#ff0000')

    assert palette.colors['red'] == '#ff0000'


def test_palette_add_color_invalid_type_name():
    '''Add a color with an invalid name type.'''
    with pytest.raises(TypeError, match='color name must be a string'):
        chromaterm.Palette().add_color(1, '#ff0000')


def test_palette_add_color_invalid_type_value():
    '''Add a color with an invalid value type.'''
    with pytest.raises(TypeError, match='color value must be a string'):
        chromaterm.Palette().add_color('red', 1)


def test_palette_add_color_invalid_value_name():
    '''Add a color with an invalid name value.'''
    with pytest.raises(ValueError, match='name accepts alphanumerics, dashes'):
        chromaterm.Palette().add_color('red$', '#ff0000')


def test_palette_add_color_invalid_value_name_duplicate():
    '''Add a color with a name that's already in use.'''
    palette = chromaterm.Palette()
    palette.add_color('red', '#ff0000')

    with pytest.raises(ValueError, match='same name already exists'):
        palette.add_color('red', '#ff0000')


def test_palette_add_color_invalid_value_name_reserved():
    '''Add a color with a reserved name.'''
    for color_type in chromaterm.COLOR_TYPES:
        with pytest.raises(ValueError, match='color name is reserved'):
            chromaterm.Palette().add_color(color_type, '#ff0000')


def test_palette_add_color_invalid_value_value():
    '''Add a color with an invalid value value.'''
    with pytest.raises(ValueError, match='must be in `#123abc` format'):
        chromaterm.Palette().add_color('red', 'h#ff0000')


def test_palette_resolve():
    '''Resolve a color name using a palette'''
    palette = chromaterm.Palette()
    palette.add_color('red', '#ff0000')
    palette.add_color('green', '#00ff00')
    palette.add_color('blue', '#0000ff')

    assert palette.resolve('b.red') == 'b#ff0000'
    assert palette.resolve('f.green bold') == 'f#00ff00 bold'
    assert palette.resolve('123 f.blue italic') == '123 f#0000ff italic'
    assert palette.resolve('f#0000ff bold') == 'f#0000ff bold'
    assert palette.resolve('random') == 'random'


def test_palette_resolve_invalid_type_name():
    '''Resolve a color name with the wrong type.'''
    with pytest.raises(TypeError, match='color must be a string'):
        chromaterm.Palette().resolve(False)


def test_palette_resolve_invalid_value_name_not_in_palette():
    '''Resolve a color name that isn't in the palette.'''
    with pytest.raises(ValueError, match='not in palette'):
        chromaterm.Palette().resolve('b.red')


def test_rule_get_matches(pcre):
    '''Get matches of rule that only colors the default regex group.'''
    color = chromaterm.Color('bold')
    rule = chromaterm.Rule('hello|world', color, pcre=pcre)

    data = b'hello world'
    expected = [(0, 5, color), (6, 11, color)]

    assert rule.get_matches(data) == expected


def test_rule_get_matches_new_line(pcre):
    '''Attempt to match a new line, which shouldn't be possible as it's considered
    as a separator (i.e. not passed to be highlighted). New lines are only used
    to replace data matched by exclusive rules.'''
    rule = chromaterm.Rule(r'hell\Wo', chromaterm.Color('bold'), pcre=pcre)

    assert not rule.get_matches(b'hell\no')


def test_rule_get_matches_groups(pcre):
    '''Get matches of rule that colors default and specific regex groups.'''
    colors = {
        0: chromaterm.Color('bold'),
        1: chromaterm.Color('b#123123'),
        2: chromaterm.Color('f#456456')
    }
    rule = chromaterm.Rule('(h)(i)', colors, pcre=pcre)

    data = b'hi'
    expected = [(0, 2, colors[0]), (0, 1, colors[1]), (1, 2, colors[2])]

    assert rule.get_matches(data) == expected


def test_rule_get_matches_groups_optional_no_matches(pcre):
    '''Get matches of rule that has a specific regex group that is optional. When
    the optional group is not in the match, the color should not be inserted.'''
    color = chromaterm.Color('bold')
    rule = chromaterm.Rule('hello(,)? world', {1: color}, pcre=pcre)

    data = b'hello, world'
    expected = [(5, 6, color)]

    assert rule.get_matches(data) == expected

    data = b'hello world'
    expected = []

    assert rule.get_matches(data) == expected


def test_rule_get_matches_no_colors(pcre):
    '''Get matches of rule that has no colors – nothing is changed.'''
    assert not chromaterm.Rule('hello', pcre=pcre).get_matches(b'hello')


def test_rule_get_matches_zero_length_match(pcre):
    '''Get matches of rule that has a regex that matches zero-length data.'''
    color = chromaterm.Color('bold')
    rule = chromaterm.Rule('hello() world', {1: color}, pcre=pcre)

    assert not rule.get_matches(b'hello world')


def test_rule_set_color_clear_existing(pcre):
    '''Clear the color of a rule by setting it to None.'''
    color = chromaterm.Color('bold')
    colors = {0: color, 1: color}
    rule = chromaterm.Rule('(hello|world)', colors, pcre=pcre)

    assert list(rule.colors) == [0, 1]
    rule.set_color(None, group=0)
    assert list(rule.colors) == [1]


def test_rule_set_color_default_group(pcre):
    '''Add to a rule a color to the default regex group (0).'''
    rule = chromaterm.Rule('hello', pcre=pcre)
    assert rule.color is None

    color = chromaterm.Color('bold')
    rule.set_color(color)
    assert rule.color is color
    assert rule.colors.get(0) is color


def test_rule_set_color_named_group(pcre):
    '''Reference a group by name. It should be converted to an index.'''
    rule = chromaterm.Rule('(?P<hi>hello)', pcre=pcre)
    assert rule.color is None

    color = chromaterm.Color('bold')
    rule.set_color(color, group='hi')
    assert rule.color is None
    assert rule.colors.get(1) is color

    rule.set_color(color)
    assert list(rule.colors) == [0, 1]


def test_rule_set_color_specific_group(pcre):
    '''Add to a rule a color to a specific regex group. Additionally, ensure
    that the dict keys are ordered according to the group number.'''
    rule = chromaterm.Rule(r'he(llo)\1', pcre=pcre)
    assert rule.color is None

    color = chromaterm.Color('bold')
    rule.set_color(color, group=1)
    assert rule.color is None
    assert rule.colors.get(1) is color

    rule.set_color(color)
    assert list(rule.colors) == [0, 1]


def test_rule_set_color_invalid_type_color(pcre):
    '''Add to a rule a color with invalid group type.'''
    rule = chromaterm.Rule('hello', pcre=pcre)

    with pytest.raises(TypeError, match='color must be a chromaterm.Color'):
        rule.set_color(True)


def test_rule_set_color_invalid_type_group(pcre):
    '''Add to a rule a color with invalid color type.'''
    rule = chromaterm.Rule('hello', pcre=pcre)

    with pytest.raises(TypeError, match='group must be an integer or a strin'):
        rule.set_color(chromaterm.Color('bold'), group=1.1)


def test_rule_set_color_invalid_value_group_index(pcre):
    '''Reference the index of a group that doesn't exist.'''
    rule = chromaterm.Rule(r'he(llo)\1', pcre=pcre)

    with pytest.raises(ValueError, match='regex has 1 group'):
        rule.set_color(chromaterm.Color('bold'), group=2)


def test_rule_set_color_invalid_value_group_name(pcre):
    '''Reference the name of a group that doesn't exist.'''
    rule = chromaterm.Rule('(?P<hi>hello)', pcre=pcre)

    with pytest.raises(ValueError, match='named group .+ not in regex'):
        rule.set_color(chromaterm.Color('bold'), group='bla')


def test_rule_change_color(pcre):
    '''Confirm that a color change overrides the old one.'''
    rule = chromaterm.Rule('hello', chromaterm.Color('bold'), pcre=pcre)

    old_color = rule.color
    rule.color = chromaterm.Color('b#123123')
    assert old_color is not rule.color


def test_rule_change_pcre(pcre):
    '''Confirm that a change in the PCRE flag recompiles the regex.'''
    # pylint: disable=protected-access
    rule = chromaterm.Rule('hello', chromaterm.Color('bold'), pcre=pcre)

    old_regex_object = rule._regex_object
    rule.pcre = not rule.pcre
    assert old_regex_object is not rule._regex_object


def test_rule_format_color_dict(pcre):
    '''Accept a dictionary of groups and colors.'''
    colors = {1: chromaterm.Color('bold'), 2: chromaterm.Color('italic')}
    rule = chromaterm.Rule('h(e)(llo)', colors, pcre=pcre)

    assert rule.colors[1] is colors[1]
    assert rule.colors[2] is colors[2]


def test_rule_invalid_type_color(pcre):
    '''Rule with an invalid color type.'''
    with pytest.raises(TypeError, match='color must be a chromaterm.Color'):
        chromaterm.Rule('hello', True, pcre=pcre)


def test_rule_invalid_type_description(pcre):
    '''Rule with an invalid description type.'''
    with pytest.raises(TypeError, match='description must be a string'):
        chromaterm.Rule('hello', description=True, pcre=pcre)


def test_rule_invalid_type_exclusive(pcre):
    '''Rule with an invalid exclusive type.'''
    with pytest.raises(TypeError, match='exclusive must be a boolean'):
        chromaterm.Rule('hello', exclusive='', pcre=pcre)


def test_rule_invalid_type_pcre():
    '''Rule with an invalid pcre type.'''
    with pytest.raises(TypeError, match='pcre must be a boolean'):
        chromaterm.Rule('hello', pcre='')


def test_rule_invalid_type_regex(pcre):
    '''Rule with an invalid regex type.'''
    with pytest.raises(TypeError, match='regex must be a string'):
        chromaterm.Rule(True, pcre=pcre)


def test_config_highlight(pcre):
    '''Highlight with one rule.'''
    rule = chromaterm.Rule('hello', chromaterm.Color('b#123123'), pcre=pcre)
    config = chromaterm.Config()
    config.rules.append(rule)

    data = b'hello world'
    expected = [
        rule.color.color_code, b'hello', rule.color.color_reset, b' world'
    ]

    assert config.highlight(data) == b''.join(expected)


def test_config_highlight_adjoin_type_different(pcre):
    '''Two rules with different color types, and one ending where the other
    starts. Both are applied without any overlap in the codes, independent of the
    order.
    1: -------
    2:        -------'''
    rule1 = chromaterm.Rule('he', chromaterm.Color('b#123123'), pcre=pcre)
    rule2 = chromaterm.Rule('llo', chromaterm.Color('f#456456'), pcre=pcre)
    config = chromaterm.Config()
    config.rules.append(rule1)
    config.rules.append(rule2)

    data = b'hello'
    expected = [
        rule1.color.color_code, b'he', rule1.color.color_reset,
        rule2.color.color_code, b'llo', rule2.color.color_reset
    ]

    assert config.highlight(data) == b''.join(expected)
    config.rules.reverse()
    assert config.highlight(data) == b''.join(expected)


def test_config_highlight_adjoin_type_mixed(pcre):
    '''Two rules with mixed color types, with one ending where the other starts.
    Both are applied without any overlap in the codes, independent of the
    order.
    1: -------
    2:        -------'''
    rule1 = chromaterm.Rule('he',
                            chromaterm.Color('b#123123 blink'),
                            pcre=pcre)
    rule2 = chromaterm.Rule('llo',
                            chromaterm.Color('b#456456 bold'),
                            pcre=pcre)
    config = chromaterm.Config()
    config.rules.append(rule1)
    config.rules.append(rule2)

    data = b'hello'
    expected = [
        rule1.color.color_code, b'he', rule1.color.color_reset,
        rule2.color.color_code, b'llo', rule2.color.color_reset
    ]

    assert config.highlight(data) == b''.join(expected)
    config.rules.reverse()
    assert config.highlight(data) == b''.join(expected)


def test_config_highlight_adjoin_type_same(pcre):
    '''Two rules with same color type, with one ending where the other starts.
    Both are applied without any overlap in the codes, independent of the order.
    1: -------
    2:        -------'''
    rule1 = chromaterm.Rule('he', chromaterm.Color('b#123123'), pcre=pcre)
    rule2 = chromaterm.Rule('llo', chromaterm.Color('b#456456'), pcre=pcre)
    config = chromaterm.Config()
    config.rules.append(rule1)
    config.rules.append(rule2)

    data = b'hello'
    expected = [
        rule1.color.color_code, b'he', rule1.color.color_reset,
        rule2.color.color_code, b'llo', rule2.color.color_reset
    ]

    assert config.highlight(data) == b''.join(expected)
    config.rules.reverse()
    assert config.highlight(data) == b''.join(expected)


def test_config_highlight_common_beginning_type_different(pcre):
    '''Two rules with different color types, and both sharing the same start of
    a match. The most recent rule will be closer to the match's start.
    1: --------------
    2: -------'''
    rule1 = chromaterm.Rule('hello', chromaterm.Color('b#123123'), pcre=pcre)
    rule2 = chromaterm.Rule('he', chromaterm.Color('f#456456'), pcre=pcre)
    config = chromaterm.Config()
    config.rules.append(rule1)
    config.rules.append(rule2)

    data = b'hello'
    expected = [
        rule1.color.color_code, rule2.color.color_code, b'he',
        rule2.color.color_reset, b'llo', rule1.color.color_reset
    ]

    assert config.highlight(data) == b''.join(expected)
    config.rules.reverse()
    expected[0], expected[1] = expected[1], expected[0]
    assert config.highlight(data) == b''.join(expected)


def test_config_highlight_common_beginning_type_mixed(pcre):
    '''Two rules with different color types, and both sharing the same start of
    a match. The most recent rule will be closer to the match's start.
    1: --------------
    2: -------'''
    rule1 = chromaterm.Rule('hello',
                            chromaterm.Color('b#123123 italic'),
                            pcre=pcre)
    rule2 = chromaterm.Rule('he', chromaterm.Color('b#456456 bold'), pcre=pcre)
    config = chromaterm.Config()
    config.rules.append(rule1)
    config.rules.append(rule2)

    data = b'hello'
    expected = [
        rule1.color.color_code, rule2.color.color_code, b'he', RESET_BOLD,
        rule1.color.color_types[0][1], b'llo', rule1.color.color_reset
    ]

    assert config.highlight(data) == b''.join(expected)


def test_config_highlight_common_beginning_type_same(pcre):
    '''Two rules with same color type, and both sharing the same start of a
    match. The most recent rule will be closer to the match's start.
    1: --------------
    2: -------'''
    rule1 = chromaterm.Rule('hello', chromaterm.Color('b#123123'), pcre=pcre)
    rule2 = chromaterm.Rule('he', chromaterm.Color('b#456456'), pcre=pcre)
    config = chromaterm.Config()
    config.rules.append(rule1)
    config.rules.append(rule2)

    data = b'hello'
    expected = [
        rule1.color.color_code, rule2.color.color_code, b'he',
        rule1.color.color_code, b'llo', rule1.color.color_reset
    ]

    assert config.highlight(data) == b''.join(expected)


def test_config_highlight_common_end_type_different(pcre):
    '''Two rules with different color types, and both sharing the same end of a
    match. The most recent rule will be closer to the match's end.
    1: --------------
    2:        -------'''
    rule1 = chromaterm.Rule('hello', chromaterm.Color('b#123123'), pcre=pcre)
    rule2 = chromaterm.Rule('llo', chromaterm.Color('f#456456'), pcre=pcre)
    config = chromaterm.Config()
    config.rules.append(rule1)
    config.rules.append(rule2)

    data = b'hello'
    expected = [
        rule1.color.color_code, b'he', rule2.color.color_code, b'llo',
        rule2.color.color_reset, rule1.color.color_reset
    ]

    assert config.highlight(data) == b''.join(expected)
    config.rules.reverse()
    expected[-1], expected[-2] = expected[-2], expected[-1]
    assert config.highlight(data) == b''.join(expected)


def test_config_highlight_common_end_type_mixed(pcre):
    '''Two rules with different color types, and both sharing the same end of a
    match. The most recent rule will be closer to the match's end.
    1: --------------
    2:        -------'''
    rule1 = chromaterm.Rule('hello',
                            chromaterm.Color('b#123123 italic'),
                            pcre=pcre)
    rule2 = chromaterm.Rule('llo',
                            chromaterm.Color('b#456456 bold'),
                            pcre=pcre)
    config = chromaterm.Config()
    config.rules.append(rule1)
    config.rules.append(rule2)

    data = b'hello'
    expected = [
        rule1.color.color_code, b'he', rule2.color.color_code, b'llo',
        RESET_BOLD, rule1.color.color_types[0][1], rule1.color.color_reset
    ]

    assert config.highlight(data) == b''.join(expected)


def test_config_highlight_common_end_type_same(pcre):
    '''Two rules with same color type, and both sharing the same end of a
    match. The most recent rule will be closer to the match's end.
    1: --------------
    2:        -------'''
    rule1 = chromaterm.Rule('hello', chromaterm.Color('b#123123'), pcre=pcre)
    rule2 = chromaterm.Rule('llo', chromaterm.Color('b#456456'), pcre=pcre)
    config = chromaterm.Config()
    config.rules.append(rule1)
    config.rules.append(rule2)

    data = b'hello'
    expected = [
        rule1.color.color_code, b'he', rule2.color.color_code, b'llo',
        rule1.color.color_code, rule1.color.color_reset
    ]

    assert config.highlight(data) == b''.join(expected)


def test_config_highlight_encapsulate_type_different(pcre):
    '''Three rules, with matches that encapsulate each other, different color
    types. None of them should affect the rest as they are of different types. The
    order in which the rules are applied should not matter.
    1:     --
    2:   ------
    3: ----------'''
    rule1 = chromaterm.Rule('lo wo', chromaterm.Color('b#123123'), pcre=pcre)
    rule2 = chromaterm.Rule('llo wor', chromaterm.Color('f#456456'), pcre=pcre)
    rule3 = chromaterm.Rule('hello world', chromaterm.Color('bold'), pcre=pcre)
    config = chromaterm.Config()
    config.rules.append(rule1)
    config.rules.append(rule2)
    config.rules.append(rule3)

    data = b'hello world'
    expected = [
        rule3.color.color_code, b'he', rule2.color.color_code, b'l',
        rule1.color.color_code, b'lo wo', rule1.color.color_reset, b'r',
        rule2.color.color_reset, b'ld', rule3.color.color_reset
    ]

    assert config.highlight(data) == b''.join(expected)
    config.rules.reverse()
    assert config.highlight(data) == b''.join(expected)


def test_config_highlight_encapsulate_type_mixed(pcre):
    '''Three rules, with matches that encapsulate each other, mixed color types.
    The colors that are different should not affect each other, but those that
    are the same should correctly track each others colors and update their
    resets appropriately.
    1:     --
    2:   ------
    3: ----------'''
    rule1 = chromaterm.Rule('lo wo',
                            chromaterm.Color('b#123123 italic'),
                            pcre=pcre)
    rule2 = chromaterm.Rule('llo wor',
                            chromaterm.Color('b#456456 bold'),
                            pcre=pcre)
    rule3 = chromaterm.Rule('hello world', chromaterm.Color('bold'), pcre=pcre)
    config = chromaterm.Config()
    config.rules.append(rule1)
    config.rules.append(rule2)
    config.rules.append(rule3)

    data = b'hello world'
    expected = [
        rule3.color.color_code, b'he', rule2.color.color_code, b'l',
        rule1.color.color_code, b'lo wo', RESET_ITALIC,
        rule2.color.color_types[0][1], b'r', rule2.color.color_types[1][1],
        chromaterm.COLOR_TYPES[rule2.color.color_types[0][0]]['reset'], b'ld',
        rule3.color.color_reset
    ]

    assert config.highlight(data) == b''.join(expected)
    config.rules.reverse()
    assert config.highlight(data) == b''.join(expected)


def test_config_highlight_encapsulate_type_same(pcre):
    '''Three rules, with matches that encapsulate each other, same color type.
    The reset of a match would be updated to the color of the encapsulating one.
    Once the second match updates the reset of the first one to a color, it is no
    longer considered a reset, and therefore the third match would not update it
    to its own color. The order in which the rules are applied should not matter.
    1:     --
    2:   ------
    3: ----------'''
    rule1 = chromaterm.Rule('lo wo', chromaterm.Color('b#123123'), pcre=pcre)
    rule2 = chromaterm.Rule('llo wor', chromaterm.Color('b#456456'), pcre=pcre)
    rule3 = chromaterm.Rule('hello world',
                            chromaterm.Color('b#abcabc'),
                            pcre=pcre)
    config = chromaterm.Config()
    config.rules.append(rule1)
    config.rules.append(rule2)
    config.rules.append(rule3)

    data = b'hello world'
    expected = [
        rule3.color.color_code, b'he', rule2.color.color_code, b'l',
        rule1.color.color_code, b'lo wo', rule2.color.color_code, b'r',
        rule3.color.color_code, b'ld', rule3.color.color_reset
    ]

    assert config.highlight(data) == b''.join(expected)
    config.rules.reverse()
    assert config.highlight(data) == b''.join(expected)


def test_config_highlight_exclusive(pcre):
    '''Two exclusive rules, partially overlapping matches. The first rule to match
    the input gets to color it.
    1:     ----------
    2: ----------'''
    rule1 = chromaterm.Rule('llo',
                            chromaterm.Color('bold'),
                            exclusive=True,
                            pcre=pcre)
    rule2 = chromaterm.Rule('hell',
                            chromaterm.Color('italic'),
                            exclusive=True,
                            pcre=pcre)
    config = chromaterm.Config()
    config.rules.append(rule1)
    config.rules.append(rule2)

    data = b'hello'
    expected = [b'he', rule1.color.color_code, b'llo', rule1.color.color_reset]

    assert config.highlight(data) == b''.join(expected)
    config.rules.reverse()
    expected = [rule2.color.color_code, b'hell', rule2.color.color_reset, b'o']
    assert config.highlight(data) == b''.join(expected)


def test_config_highlight_overlap_full_type_different(pcre):
    '''Two rules, fully overlapping matches, different color types. Should not
    affect each other as they have different types. Most recent rule should be
    closest to the match.
    1: ----------
    2: ----------'''
    rule1 = chromaterm.Rule('hello', chromaterm.Color('b#123123'), pcre=pcre)
    rule2 = chromaterm.Rule('hello', chromaterm.Color('f#456456'), pcre=pcre)
    config = chromaterm.Config()
    config.rules.append(rule1)
    config.rules.append(rule2)

    data = b'hello'
    expected = [
        rule1.color.color_code, rule2.color.color_code, b'hello',
        rule2.color.color_reset, rule1.color.color_reset
    ]

    assert config.highlight(data) == b''.join(expected)


def test_config_highlight_overlap_full_type_mixed(pcre):
    '''Two rules, fully overlapping matches, mixed color types. The different
    color types should not affect each other, but those that are the same should
    update the oldest reset with the most recent rule's color code.
    1: ----------
    2: ----------'''
    rule1 = chromaterm.Rule('hello',
                            chromaterm.Color('b#123123 bold'),
                            pcre=pcre)
    rule2 = chromaterm.Rule('hello',
                            chromaterm.Color('b#456456 italic'),
                            pcre=pcre)
    config = chromaterm.Config()
    config.rules.append(rule1)
    config.rules.append(rule2)

    data = b'hello'
    expected = [
        rule1.color.color_code, rule2.color.color_code, b'hello', RESET_ITALIC,
        rule1.color.color_types[0][1], rule1.color.color_reset
    ]

    assert config.highlight(data) == b''.join(expected)


def test_config_highlight_overlap_full_type_same(pcre):
    '''Two rules, fully overlapping matches, same color type. The reset of the
    first rule should be replaced with the color code of the second (most recent)
    rule to prevent the reset from interrupting the color.
    1: ----------
    2: ----------'''
    rule1 = chromaterm.Rule('hello', chromaterm.Color('b#123123'), pcre=pcre)
    rule2 = chromaterm.Rule('hello', chromaterm.Color('b#456456'), pcre=pcre)
    config = chromaterm.Config()
    config.rules.append(rule1)
    config.rules.append(rule2)

    data = b'hello'
    expected = [
        rule1.color.color_code, rule2.color.color_code, b'hello',
        rule1.color.color_code, rule1.color.color_reset
    ]

    assert config.highlight(data) == b''.join(expected)


def test_config_highlight_overlap_partial_type_different(pcre):
    '''Two rules, partially overlapping matches, different color types. The rules
    should not affect each other. Tested in reverse order, too.
    1: ----------
    2:     ----------'''
    rule1 = chromaterm.Rule('hell', chromaterm.Color('b#123123'), pcre=pcre)
    rule2 = chromaterm.Rule('llo', chromaterm.Color('f#456456'), pcre=pcre)
    config = chromaterm.Config()
    config.rules.append(rule1)
    config.rules.append(rule2)

    data = b'hello'
    expected = [
        rule1.color.color_code, b'he', rule2.color.color_code, b'll',
        rule1.color.color_reset, b'o', rule2.color.color_reset
    ]

    assert config.highlight(data) == b''.join(expected)
    config.rules.reverse()
    assert config.highlight(data) == b''.join(expected)


def test_config_highlight_overlap_partial_type_mixed(pcre):
    '''Two rules, partially overlapping matches, mixed color types. The different
    color types should not affect each other, but those that are the same should
    correctly update the reset that is in the middle of the other match to that
    of the match's color. Order of rules should not matter.
    1: ----------
    2:     ----------'''
    rule1 = chromaterm.Rule('hell',
                            chromaterm.Color('b#123123 bold'),
                            pcre=pcre)
    rule2 = chromaterm.Rule('llo',
                            chromaterm.Color('b#456456 italic'),
                            pcre=pcre)
    config = chromaterm.Config()
    config.rules.append(rule1)
    config.rules.append(rule2)

    data = b'hello'
    expected = [
        rule1.color.color_code, b'he', rule2.color.color_code, b'll',
        RESET_BOLD, rule2.color.color_types[0][1], b'o',
        rule2.color.color_reset
    ]

    assert config.highlight(data) == b''.join(expected)
    config.rules.reverse()
    assert config.highlight(data) == b''.join(expected)


def test_config_highlight_overlap_partial_type_same(pcre):
    '''Two rules, partially overlapping matches, same color type. The reset of
    the first rule should be replaced with the color code of the second (most
    recent) rule to prevent the reset from interrupting the color. The behavior
    should be consistent regardless of the order of the rules.
    1: ----------
    2:     ----------'''
    rule1 = chromaterm.Rule('hell',
                            chromaterm.Color('b#123123 bold'),
                            pcre=pcre)
    rule2 = chromaterm.Rule('llo',
                            chromaterm.Color('b#456456 bold'),
                            pcre=pcre)
    config = chromaterm.Config()
    config.rules.append(rule1)
    config.rules.append(rule2)

    data = b'hello'
    expected = [
        rule1.color.color_code, b'he', rule2.color.color_code, b'll',
        rule2.color.color_types[1][1], rule2.color.color_types[0][1], b'o',
        rule2.color.color_reset
    ]

    assert config.highlight(data) == b''.join(expected)
    config.rules.reverse()
    assert config.highlight(data) == b''.join(expected)


def test_config_highlight_no_rules():
    '''Highlight with config that has no rules – nothing is changed.'''
    config = chromaterm.Config()
    assert config.highlight(b'hello world') == b'hello world'


def test_config_highlight_tracking_common_beginning_type_different(pcre):
    '''A rule with a match that has a color of a different type just before its
    start. The rule's color is closer to the match and the reset is unaffected by
    the existing color.
    1: x-------------'''
    rule = chromaterm.Rule('hello', chromaterm.Color('b#123123'), pcre=pcre)
    config = chromaterm.Config()
    config.rules.append(rule)

    data = b'\x1b[33mhello'
    expected = [
        b'\x1b[33m', rule.color.color_code, b'hello', rule.color.color_reset
    ]

    assert config.highlight(data) == b''.join(expected)


def test_config_highlight_tracking_common_beginning_type_same(pcre):
    '''A rule with a match that has a color of the same type just before its
    start. The rule's color is closer to the match and the reset used is the
    existing color.
    1: x-------------'''
    rule = chromaterm.Rule('hello', chromaterm.Color('f#456456'), pcre=pcre)
    config = chromaterm.Config()
    config.rules.append(rule)

    data = b'\x1b[33mhello'
    expected = [b'\x1b[33m', rule.color.color_code, b'hello', b'\x1b[33m']

    assert config.highlight(data) == b''.join(expected)


def test_config_highlight_tracking_common_end_type_different(pcre):
    '''A rule with a match that has a color of a different type just after its
    end. The rule's reset is closer to the match and the reset is unaffected by
    the existing color.
    1: -------------x'''
    rule = chromaterm.Rule('hello', chromaterm.Color('b#123123'), pcre=pcre)
    config = chromaterm.Config()
    config.rules.append(rule)

    data = b'hello\x1b[33m'
    expected = [
        rule.color.color_code, b'hello', rule.color.color_reset, b'\x1b[33m'
    ]

    assert config.highlight(data) == b''.join(expected)


def test_config_highlight_tracking_common_end_type_same(pcre):
    '''A rule with a match that has a color of the same type just after its end.
    The rule's reset is closer to the match and is unaffected by the existing
    color.
    1: -------------x'''
    rule = chromaterm.Rule('hello', chromaterm.Color('f#456456'), pcre=pcre)
    config = chromaterm.Config()
    config.rules.append(rule)

    data = b'hello\x1b[33m'
    expected = [
        rule.color.color_code, b'hello', rule.color.color_reset, b'\x1b[33m'
    ]

    assert config.highlight(data) == b''.join(expected)


def test_config_highlight_tracking_full_reset_beginning(pcre):
    '''A rule with a match that has a full reset just before the start of the
    match. The rule's color is closer to the match and the reset used is the
    default for that color type.
    1: R-------------'''
    rule = chromaterm.Rule('hello', chromaterm.Color('f#456456'), pcre=pcre)
    config = chromaterm.Config()
    config.rules.append(rule)

    data = b'\x1b[0mhello'
    expected = [
        b'\x1b[0m', rule.color.color_code, b'hello',
        chromaterm.COLOR_TYPES[rule.color.color_types[0][0]]['reset']
    ]

    assert config.highlight(data) == b''.join(expected)


def test_config_highlight_tracking_full_reset_end(pcre):
    '''A rule with a match that has a full reset just after the end of the match.
    The rule's reset is closer to the match and the reset used is the default for
    that color type.
    1: -------------R'''
    rule = chromaterm.Rule('hello', chromaterm.Color('f#456456'), pcre=pcre)
    config = chromaterm.Config()
    config.rules.append(rule)

    data = b'hello\x1b[0m'
    expected = [
        rule.color.color_code, b'hello',
        chromaterm.COLOR_TYPES[rule.color.color_types[0][0]]['reset'],
        b'\x1b[0m'
    ]

    assert config.highlight(data) == b''.join(expected)


def test_config_highlight_tracking_full_reset_middle(pcre):
    '''A rule with a match that has a full reset in the middle of it. The full
    reset is changed to the color code of the match and the reset of the match
    is changed to a full reset.
    1: ------R-------'''
    rule = chromaterm.Rule('hello', chromaterm.Color('f#456456'), pcre=pcre)
    config = chromaterm.Config()
    config.rules.append(rule)

    data = b'hel\x1b[0mlo'
    expected = [
        rule.color.color_code, b'hel', rule.color.color_code, b'lo', b'\x1b[0m'
    ]

    assert config.highlight(data) == b''.join(expected)


def test_config_highlight_tracking_malformed(pcre):
    '''A rule with a match that has a malformed SGR in the middle. It should be
    ignored and inserted back into the match. Highlighting from the rule should
    still go through.
    1: x-------------'''
    rule = chromaterm.Rule('hello', chromaterm.Color('b#123123'), pcre=pcre)
    config = chromaterm.Config()
    config.rules.append(rule)

    data = b'he\x1b[38;5mllo'
    expected = [
        rule.color.color_code, b'he', b'\x1b[38;5m', b'llo',
        rule.color.color_reset
    ]

    assert config.highlight(data) == b''.join(expected)


def test_config_highlight_tracking_mixed_full_reset(pcre):
    '''Track multiple color types and ensure a full reset only defaults the types
    that were not updated by other colors in the data.'''
    rule1 = chromaterm.Rule('hello', chromaterm.Color('f#456456'), pcre=pcre)
    rule2 = chromaterm.Rule('world', chromaterm.Color('b#123123'), pcre=pcre)
    config = chromaterm.Config()
    config.rules.append(rule1)
    config.rules.append(rule2)

    data = b'\x1b[33mhello\x1b[0m there \x1b[43mworld'
    expected = [
        b'\x1b[33m', rule1.color.color_code, b'hello', b'\x1b[33m', b'\x1b[0m',
        b' there ', b'\x1b[43m', rule2.color.color_code, b'world', b'\x1b[43m'
    ]

    assert config.highlight(data) == b''.join(expected)

    # The color of rule1 was reset to its default because a full reset came after
    # it, but the color of rule2 was already updated so it wasn't affected by the
    # full reset
    data = b'hello there world'
    expected = [
        rule1.color.color_code, b'hello', rule1.color.color_reset, b' there ',
        rule2.color.color_code, b'world', b'\x1b[43m'
    ]

    assert config.highlight(data) == b''.join(expected)


def test_config_highlight_tracking_multiline_type_different(pcre):
    '''Ensure that data with an existing color is tracked across highlights and
    does not affect the reset of a color of a different type.'''
    rule = chromaterm.Rule('hello', chromaterm.Color('b#123123'), pcre=pcre)
    config = chromaterm.Config()
    config.rules.append(rule)

    # Inject a foreground color to have it tracked
    assert config.highlight(b'\x1b[33m') == b'\x1b[33m'

    data = b'hello'
    expected = [rule.color.color_code, b'hello', rule.color.color_reset]

    assert config.highlight(data) == b''.join(expected)


def test_config_highlight_tracking_multiline_type_same(pcre):
    '''Ensure that data with an existing color is tracked across highlights and
    affects the reset of a color of the same type.'''
    rule = chromaterm.Rule('hello', chromaterm.Color('f#456456'), pcre=pcre)
    config = chromaterm.Config()
    config.rules.append(rule)

    # Inject a foreground color to have it tracked
    assert config.highlight(b'\x1b[33m') == b'\x1b[33m'

    data = b'hello'
    expected = [rule.color.color_code, b'hello', b'\x1b[33m']

    assert config.highlight(data) == b''.join(expected)



'''Color your output to terminal'''

from .__version__ import __version__
from .core import Color, Palette, Rule, Config, COLOR_TYPES

__all__ = ['Color', 'Palette', 'Rule', 'Config', 'COLOR_TYPES', '__version__']

# Support for python -m chromaterm
if __name__ == '__main__':
    import sys
    from .cli import main
    sys.exit(main())


'''chromaterm.default_config tests'''
import os

import chromaterm.default_config


def assert_matches(positives, negatives, rule, match_count=1):
    '''Assert that all positives are matched while negatives are not.'''

    def permutate_data(data):
        output = []

        for entry in data if isinstance(data, list) else [data]:
            entry = entry.encode()

            output.append(entry)  # Plain entry
            output.append(b'hello ' + entry)  # Start changed
            output.append(entry + b' world')  # End changed
            output.append(b'hello ' + entry + b' world')  # Both changed

        return output

    for data in permutate_data(positives):
        assert len(rule.get_matches(data)) == match_count

    for data in permutate_data(negatives):
        assert len(rule.get_matches(data)) == 0


def test_rule_numbers(pcre):
    '''Default rule: Numbers.'''
    positives = ['1', '123', '123.123']
    negatives = ['1.1.', '.1.1', '1.2.3']
    rule = chromaterm.default_config.RULE_NUMBERS
    rule.pcre = pcre

    assert_matches(positives, negatives, rule)


def test_rule_url(pcre):
    '''Default rule: URL.'''
    positives = [
        'http://example', 'http://example', 'https://example', 'ftp://example',
        'ldap://example', 'ssh://example', 'telnet://example',
        'https://www.example.com', 'http://user@example', 'http://_a.example',
        'http://user:password@example', 'http://example/hello/',
        'http://example/hello/world/', 'http://example/?-+=~@%&?#.:;,()\\/'
    ]
    negatives = [
        'bob://example', 'http:/example', 'http://.example',
        'http://$.com/hello', 'http://@example.com'
    ]
    rule = chromaterm.default_config.RULE_URL
    rule.pcre = pcre

    assert_matches(positives, negatives, rule)

    # Trailing characters that can be part of the URL but don't end with it
    assert rule.get_matches(b'http://example.')[0][:2] == (0, 14)
    assert rule.get_matches(b'http://example.html')[0][:2] == (0, 19)
    assert rule.get_matches(b'http://example:')[0][:2] == (0, 14)
    assert rule.get_matches(b'http://example;')[0][:2] == (0, 14)
    assert rule.get_matches(b'http://example,')[0][:2] == (0, 14)
    assert rule.get_matches(b'http://example)')[0][:2] == (0, 14)


def test_rule_ipv4(pcre):
    '''Default rule: IPv4 addresses.'''
    positives = ['192.168.2.1', '255.255.255.255', '=240.3.2.1', '1.2.3.4/32']
    negatives = ['256.255.255.255', '1.2.3']
    rule = chromaterm.default_config.RULE_IPV4
    rule.pcre = pcre

    assert_matches(positives, negatives, rule)


def test_rule_ipv6(pcre):
    '''Default rule: IPv6 addresses.'''
    positives = [
        'A:b:3:4:5:6:7:8', 'A::', 'A:b:3:4:5:6:7::', 'A::8', '::b:3:4:5:6:7:8',
        '::8', 'A:b:3:4:5:6::8', 'A:b:3:4:5::7:8', 'A:b:3:4::6:7:8', '::',
        'A:b:3::5:6:7:8', 'A:b::4:5:6:7:8', 'A::3:4:5:6:7:8', 'A::7:8', 'A::8',
        'A:b:3:4:5::8', 'A::6:7:8', 'A:b:3:4::8', 'A::5:6:7:8', 'A:b:3::8',
        'A::4:5:6:7:8', 'A:b::8', 'A:b:3:4:5:6:7:8/64', 'fe80::1%tun', ':::'
    ]
    negatives = [
        '1:2', '1:2:3', '1:2:3:4', '1:2:3:4:5', '1:2:3:4:5:6:7', ':::1'
        '1:2:3:4:5:6:7:8:9', '1:2:3:4:5:6:7::8', 'abcd:xyz::1', 'fe80:1%tun'
    ]
    rule = chromaterm.default_config.RULE_IPV6
    rule.pcre = pcre

    assert_matches(positives, negatives, rule)


def test_rule_mac(pcre):
    '''Default rule: MAC addresses.'''
    positives = ['0A:23:45:67:89:AB', '0a:23:45:67:89:ab', '0a23.4567.89ab']
    negatives = [
        '0A:23:45:67:89', '0A:23:45:67:89:AB:', '0A23.4567.89.AB',
        '0a23.4567.89ab.', '0a:23:45:67:xy:zx', '0a23.4567.xyzx'
    ]
    rule = chromaterm.default_config.RULE_MAC
    rule.pcre = pcre

    assert_matches(positives, negatives, rule)


def test_rule_size(pcre):
    '''Default rule: Size.'''
    positives = [
        '100b', '1kb', '1kib', '1Kb', '1kbps', '1.1kb', '1mb', '1gb', '1tb',
        '1pb', '1eb', '1zb', '1yb'
    ]
    negatives = ['100', '1kg', '1kic', '1kbpm', '1kibb', '1.1.kb', '100wb']
    rule = chromaterm.default_config.RULE_SIZE
    rule.pcre = pcre

    assert_matches(positives, negatives, rule, match_count=2)


def test_rule_date(pcre):
    '''Default rule: Date.'''
    positives = [
        '2019-12-31', '2019-12-31', '2019/12/31', '12/31/2019', '31/12/2019',
        '31/12/19', '31-12-19', 'jan 2019', 'feb 2019', 'Mar 2019', 'apr 2019',
        'MAY 2019', 'Jun 2019', 'jul  2019', 'AUG 19', 'sep 20', 'oct 21',
        'dec 23', 'AUG 19 2021', '24 jan', '25 feb 2019'
    ]
    negatives = [
        '201-12-31', '2019-13-31', '2019-12-32', '2019-12/31', '2019/12-31',
        '12-31/2019', '12/31-2019', 'xyz 2019', 'Jun 201', 'xyz 26', 'jun 32',
        '32 jun'
    ]
    rule = chromaterm.default_config.RULE_DATE
    rule.pcre = pcre

    assert_matches(positives, negatives, rule)

    # MMM DD takes precedence over DD MMM, if both are matched
    assert rule.get_matches(b'12 Mar 13')[0][:2] == (3, 9)

    # The '13' should be considered part of the time (i.e. ignored)
    assert rule.get_matches(b'12 Mar  13:59')[0][:2] == (0, 6)


def test_rule_time(pcre):
    '''Default rule: Time.'''
    positives = [
        '23:59', '23:01', '23:01:01', '23:01:01.123', '23:01:01-0800',
        '23:01:01+08', '23:01:01Z'
    ]
    negatives = ['24:59', '23:60', '23:1', '23:01:1', '23:01:01:', '23:01.123']
    rule = chromaterm.default_config.RULE_TIME
    rule.pcre = pcre

    assert_matches(positives, negatives, rule)


def test_write_default_config():
    '''Write config file.'''
    path = __name__ + '1'
    assert chromaterm.default_config.write_default_config(path)
    assert os.access(path, os.F_OK)
    os.remove(path)


def test_write_default_config_exists():
    '''Config file already exists.'''
    path = __name__ + '2'
    with open(path, 'w', encoding='utf-8'):
        assert not chromaterm.default_config.write_default_config(path)
    os.remove(path)


def test_write_default_config_no_permission():
    '''No write permission on directory.'''
    path = __name__ + '3' + '/hi.yml'
    os.mkdir(os.path.dirname(path), mode=0o444)
    assert not chromaterm.default_config.write_default_config(path + '/hi')
    os.rmdir(os.path.dirname(path))



'''A PCRE2 implementation. Similar to Python's `re`, albeit very minimal.'''
import collections
import re
from ctypes import (CDLL, POINTER, byref, c_char, c_size_t, c_uint32,
                    create_string_buffer)
from ctypes.util import find_library

PCRE2 = CDLL(find_library('pcre2-8'))
PCRE2.pcre2_compile_8.restype = POINTER(c_char)
PCRE2.pcre2_match_data_create_8.restype = POINTER(c_char)

# PCRE2_NOTEMPTY_ATSTART = 4
MATCH_OPTIONS = c_uint32(4)


class GroupIndex(collections.UserDict):
    '''Extracting the named groups table is too much work. This class patches
    `__getitem__` to extract a named group on-demand (much easier to do).'''

    def __init__(self, regex):
        super().__init__()
        self._regex = regex

    def __getitem__(self, name):
        group_id = PCRE2.pcre2_substring_number_from_name_8(
            self._regex,
            name.encode(),
        )

        if group_id < 1:
            raise KeyError(name)

        return group_id

    def get(self, name, default=None):
        '''Get method to match dict interface used in core.py'''
        try:
            return self[name]
        except KeyError:
            return default


class Pattern:
    '''PCRE2 Pattern. Like that of Python's, if it was stripped of everything.'''

    def __init__(self, pattern):
        '''Constructor.

        Args:
            pattern (bytes): The regex pattern.
        '''
        error_buffer = create_string_buffer(4096)
        error_code = c_uint32()
        error_offset = c_size_t()
        groups = c_uint32()

        self.pattern = pattern
        self._regex = PCRE2.pcre2_compile_8(
            self.pattern,
            c_size_t(len(self.pattern)),
            c_uint32(0),
            byref(error_code),
            byref(error_offset),
            None,
        )

        if error_code.value != 100:
            PCRE2.pcre2_get_error_message_8(
                error_code,
                error_buffer,
                byref(c_size_t(len(error_buffer) - 1)),
            )
            msg = f'position {error_offset.value}: {error_buffer.value.decode()}'

            raise re.error(msg)

        # PCRE2_JIT_COMPLETE = 1
        PCRE2.pcre2_jit_compile_8(self._regex, c_uint32(1))

        # PCRE2_INFO_CAPTURECOUNT = 4
        PCRE2.pcre2_pattern_info_8(self._regex, c_uint32(4), byref(groups))

        self.groups = groups.value
        self.groupindex = GroupIndex(self._regex)

        self._match_data = PCRE2.pcre2_match_data_create_8(
            c_uint32(self.groups + 1), None)

        PCRE2.pcre2_get_ovector_pointer_8.restype = POINTER(
            2 * (self.groups + 1) * c_size_t)

        # Emulate `span` method of `re.Match`
        match = PCRE2.pcre2_get_ovector_pointer_8(self._match_data).contents
        match.span = lambda gid: (match[2 * gid], match[2 * gid + 1])

        self._match = match
        self._data_len = c_size_t()

    def __del__(self):
        if PCRE2:
            if hasattr(self, '_match_data'):
                PCRE2.pcre2_match_data_free_8(self._match_data)
            if hasattr(self, '_regex'):
                PCRE2.pcre2_code_free_8(self._regex)

    def finditer(self, data):
        '''A generator for finding matches.

        Args:
            data (bytes): The subject of the search.
        '''
        # Reset the startoffset to 0
        self._match[1] = 0
        self._data_len.value = len(data)

        while PCRE2.pcre2_jit_match_8(
                self._regex,
                data,
                self._data_len,
                self._match[1],
                MATCH_OPTIONS,
                self._match_data,
                None,
        ) >= 0:
            yield self._match



import os
import codecs
from setuptools import find_packages, setup

here = os.path.abspath(os.path.dirname(__file__))
readme = None

requirements = ["psutil", "PyYAML>=5.1"]

test_requirements = []

with open("README.md", "r", encoding="utf-8") as f:
    readme = f.read()

about = dict()
with codecs.open(os.path.join(here, "chromaterm", "__version__.py"), "r", "utf-8") as f:
    exec(f.read(), about)

setup(
    name="chromaterm2",
    url="https://github.com/rgcr/ChromaTerm",
    license="MIT",
    description="Color your Terminal with Regex! (Maintained fork of ChromaTerm)",
    version=about["__version__"],

    author=about["__author__"],
    author_email=about["__email__"],

    entry_points={"console_scripts": ["ct = chromaterm.cli:main"]},

    install_requires=requirements,
    setup_requires=requirements,
    tests_require=test_requirements,

    long_description=readme + "\n\n",
    long_description_content_type="text/markdown",
    packages=find_packages(),

    python_requires=">=3.9.0",
    classifiers=[
        "Intended Audience :: Information Technology",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Telecommunications Industry",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Topic :: Terminals",
        "Topic :: Utilities",
    ],
)



'''A store for the default rules of ChromaTerm'''
import os

import yaml

from .core import Color, Palette, Rule

# pylint: disable=line-too-long

PALETTE = Palette()

# https://coolors.co/c68c6c-00e0d1-ef2e9f-5698c8-a35a00-df99f0-03d28d
PALETTE.add_color('type-1', '#dc8968')
PALETTE.add_color('type-2', '#00e0d1')
PALETTE.add_color('type-3', '#ef2e9f')
PALETTE.add_color('type-4', '#5698c8')
PALETTE.add_color('type-5', '#a35a00')
PALETTE.add_color('type-6', '#df99f0')
PALETTE.add_color('type-7', '#03d28d')

# https://coolors.co/c71800-c96901-ca9102-cab902-a2bc02-79bf02-28c501
PALETTE.add_color('status-1', '#c71800')
PALETTE.add_color('status-2', '#c96901')
PALETTE.add_color('status-3', '#ca9102')
PALETTE.add_color('status-4', '#cab902')
PALETTE.add_color('status-5', '#a2bc02')
PALETTE.add_color('status-6', '#79bf02')
PALETTE.add_color('status-7', '#28c501')

RULE_NUMBERS = Rule(
    r'\b(?<!\.)\d+(\.\d+)?(?!\.)\b',
    Color('f.type-1', palette=PALETTE),
    'Numbers',
)

RULE_URL = Rule(
    r'''(?ix)\b
    ((htt|ft|lda)ps?|telnet|ssh)://  # Scheme
    ([-%:\w\\/]{1,256}@)?  # User info
    [-\w]{1,63}(\.[-\w]{1,63}){0,126}(:\d{1,5})?  # Host and port
    (/[-+=~@%&?#.:;,\w\\/()]*)?  # Path, query, and fragment
((?=[.:;,)])|\b)  # Avoid highlighting trailing path characters by matching them in a lookahead
''',
    Color('f.type-4', palette=PALETTE),
    'URL',
    exclusive=True,
)

RULE_IPV4 = Rule(
    r'\b(?<!\.)((25[0-5]|(2[0-4]|[0-1]?\d)?\d)\.){3}(25[0-5]|(2[0-4]|[0-1]?\d)?\d)(/\d+)?\b',
    Color('f.type-2', palette=PALETTE),
    'IPv4',
    exclusive=True,
)

RULE_IPV6 = Rule(
    r'''(?ix)(?<![\w:])(
    ([\da-f]{1,4}:){7}[\da-f]{1,4}|  # 1:2:3:4:5:6:7:8
    [\da-f]{1,4}:(:[\da-f]{1,4}){1,6}|  # 1::3:4:5:6:7:8
    ([\da-f]{1,4}:){1,2}(:[\da-f]{1,4}){1,5}|  # 1:2::4:5:6:7:8
    ([\da-f]{1,4}:){1,3}(:[\da-f]{1,4}){1,4}|  # 1:2:3::5:6:7:8
    ([\da-f]{1,4}:){1,4}(:[\da-f]{1,4}){1,3}|  # 1:2:3:4::6:7:8
    ([\da-f]{1,4}:){1,5}(:[\da-f]{1,4}){1,2}|  # 1:2:3:4:5::7:8
    ([\da-f]{1,4}:){1,6}:[\da-f]{1,4}|  # 1:2:3:4:5:6::8
    ([\da-f]{1,4}:){1,7}:|  # 1:2:3:4:5:6:7::
    :((:[\da-f]{1,4}){1,7}|:)  # ::2:3:4:5:6:7:8
)(:(?=\W))?  # \W is an exclusive-flag hack to color the : before an IPv4-embedded address
(%[\da-z]+)?  # Zone index
(/\d+)?  # Prefix length
(?!:?\w)
''',
    Color('f.type-3', palette=PALETTE),
    'IPv6',
    exclusive=True,
)

RULE_MAC = Rule(
    r'''(?ix)\b(
    (?<!:)([\da-f]{1,2}:){5}[\da-f]{1,2}(?!:)|  # 11:22:33:aa:bb:cc
    (?<!\.)([\da-f]{4}\.){2}[\da-f]{4}(?!\.)  # 1122.33aa.bbcc
)\b
''',
    Color('f.type-4', palette=PALETTE),
    'MAC address',
    exclusive=True,
)

RULE_DATE = Rule(
    r'''(?ix)\b(
    (\d{2}|\d{4})(?P<sep1>[-/])(0?[1-9]|1[0-2])(?P=sep1)(3[0-1]|[1-2]\d|0?[1-9])|  # YYYY-MM-DD, YY-MM-DD, YYYY/MM/DD, YY/MM/DD
    (3[0-1]|[1-2]\d|0?[1-9])(?P<sep2>[-/])(0?[1-9]|1[0-2])(?P=sep2)(\d{2}|\d{4})|  # DD-MM-YYYY, DD-MM-YY, DD/MM/YYYY, DD/MM/YY
    (0?[1-9]|1[0-2])(?P<sep3>[-/])(3[0-1]|[1-2]\d|0?[1-9])(?P=sep3)(\d{2}|\d{4})|  # MM-DD-YYYY, MM-DD-YY, MM/DD/YYYY, MM/DD/YY
    (jan|feb|mar|apr|may|jun|jul|aug|sep|oct|nov|dec)\s+(  # MMM
        (3[0-1]|[1-2]\d|0?[1-9])(\s+\d{4})?|\d{4}  # DD (YYYY)?, YYYY
    )|(3[0-1]|[1-2]\d|0?[1-9])\s(jan|feb|mar|apr|may|jun|jul|aug|sep|oct|nov|dec)(?!\s+(3[0-1]|[1-2]\d|0?[1-9])([^\w:]|$))(\s+\d{4})?  # DD MMM (YYYY)?
)((?=[\WT_])|$)
''',
    Color('b.type-5', palette=PALETTE),
    'Date',
    exclusive=True,
)

RULE_TIME = Rule(
    r'''(?ix)(?<![\.:])(\b|(?<=T))
    (2[0-3]|[0-1]\d):[0-5]\d  # Hours and minutes
    (:[0-5]\d([\.,]\d{3,6})?)?  # (Seconds (sub-seconds, 3 to 6 decimal places)?)?
    ([\-\+](\d{2}|\d{4})|Z)?  # (Timezone)?
(?![\.:])\b
''',
    Color('b.type-5', palette=PALETTE),
    'Time',
    exclusive=True,
)

RULE_SIZE = Rule(
    r'(?i)\b\d+(\.\d+)?\s?((([KMGTPEZY](i?B)?)|B)(ps)?)\b',
    {
        0: Color('f.type-6', palette=PALETTE),
        2: Color('bold'),
    },
    'Size, like 123G 123Gb 123Gib 1.23G 123Gbps',
    exclusive=True,
)

RULE_GENERIC_BAD = Rule(
    r'(?i)\b(password|abnormal(ly)?|down|los(t|ing)|err(ors?)?|(den(y|ies|ied)?)|reject(ing|ed)?|drop(ped|s)?|(err\-)?disabled?|(timed?\-?out)|fail(s|ed|iure)?|disconnect(ed)?|unreachable|invalid|bad|notconnect|unusable|blk|inaccessible|wrong|collisions?|unsynchronized|mismatch|runts)\b',
    Color('f.status-1', palette=PALETTE),
    'Generic - Bad',
)

RULE_GENERIC_AMBIGIOUS_BAD = Rule(
    r'(?i)\b(no(pe)?|exit(ed)?|reset(t?ing)?|discard(ed|ing)?|block(ed|ing)?|filter(ed|ing)?|stop(p(ed|ing))?|never|bad)\b',
    Color('f.status-3', palette=PALETTE),
    'Generic - Ambigious bad',
)

RULE_GENERIC_NOT_TOO_BAD = Rule(
    r'(?i)\b(warnings?)\b',
    Color('f.status-4', palette=PALETTE),
    'Generic - Not too bad',
)

RULE_GENERIC_AMBIGIOUS_GOOD = Rule(
    r'(?i)\b(ye(s|ah?|p)?|start(ed|ing)?|running|good)\b',
    Color('f.status-6', palette=PALETTE),
    'Generic - Ambigious good',
)

RULE_GENERIC_GOOD = Rule(
    r'(?i)\b(up|ok(ay)?|permit(ed|s)?|accept(s|ed)?|enabled?|online|succe((ss(ful|fully)?)|ed(ed)?)?|connect(ed)?|reachable|valid|forwarding|synchronized)\b',
    Color('f.status-7', palette=PALETTE),
    'Generic - Good',
)


def generate_default_rules_yaml():
    '''Returns a YAML string of the default configuration.'''
    data = yaml.dump({'palette': PALETTE.colors}, sort_keys=False) + '\n'
    data += 'rules:\n'

    for rule in [
            RULE_NUMBERS, RULE_URL, RULE_IPV4, RULE_IPV6, RULE_MAC, RULE_DATE,
            RULE_TIME, RULE_SIZE, RULE_GENERIC_BAD, RULE_GENERIC_AMBIGIOUS_BAD,
            RULE_GENERIC_NOT_TOO_BAD, RULE_GENERIC_AMBIGIOUS_GOOD,
            RULE_GENERIC_GOOD
    ]:
        rule_dict = {'description': rule.description, 'regex': rule.regex}

        if list(rule.colors) == [0]:
            rule_dict['color'] = rule.color.color
        else:
            rule_dict['color'] = {k: v.color for k, v in rule.colors.items()}

        if rule.exclusive:
            rule_dict['exclusive'] = True

        data += yaml.dump([rule_dict], sort_keys=False) + '\n'

    return data


def write_default_config(path):
    '''Writes a default config file if it doesn't exist.

    Args:
        path (str): The location to which the default config file is written.

    Returns:
        True if the file did not exist and was written. False otherwise.
    '''
    # Already exists
    if os.access(path, os.F_OK):
        return False

    # No write permission in directory
    if not os.access(os.path.dirname(path) or os.path.curdir, os.W_OK):
        return False

    with open(path, 'w', encoding='utf-8') as file:
        file.write(generate_default_rules_yaml())

    return True


def yaml_str_presenter(dumper, data):
    '''YAML string representer that uses the `|` style when data is multiline.'''
    style = '|' if len(data.splitlines()) > 1 else None
    return dumper.represent_scalar('tag:yaml.org,2002:str', data, style=style)


yaml.add_representer(str, yaml_str_presenter)
